class AuthenticationError(BaseException):
    def __init__(self, msg: str):
        __str__ = msg
        return None

class ReadingMapFileError(BaseException):
    def __init__(self, msg: str):
        __str__ = msg
        return None

class EmptyMapFileError(BaseException):
    def __init__(self, msg: str):
        __str__ = msg
        return None

class ReportCreationError(BaseException):
    def __init__(self, msg: str):
        __str__ = msg
        return None


from fore_cloudreach.auth import Auth
from fore_cloudreach.template import Template
from fore_cloudreach.ingester import Ingester
from fore_cloudreach.errors import AuthenticationError, ReadingMapFileError, EmptyMapFileError, ReportCreationError
#import pkg_resources
from importlib.metadata import version

token = None
gcreds = None
debug_mode = False

def switch_debug_mode(mode=False) -> bool:
    """
    **debug_mode** control the logging verbosity 

    Args:
        mode: boolean
        If missing the default value is False. Set the current value for the global variable `debug_mode`;

    returns:
        boolean 
        Shows the current mode: True the debug logging is ON and False for debug logging is OFF;
    """

    global debug_mode 
    debug_mode = mode
    print(f"The debug logging mode set to: {debug_mode}")

    return  debug_mode


def get_package_version():
    """
    Returning the current version of the package
    """
    v = version('fore_cloudreach')
    print(f"The package 'fore_cloudreach' version: {v}")
    
    return v



import ast
import sys
import fore_cloudreach as fc
from fore_cloudreach.errors import AuthenticationError
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
import json
import boto3
from botocore.exceptions import ClientError

SCOPES = ['https://www.googleapis.com/auth/spreadsheets'] 
 
class Auth:
    """
    The Class `Auth` will authenticate the current user with its Google's account
    """

    def __init__(self, aws_session: object, aws_credentials: list):
        """
        This will instantiate an object from the class Auth.
        The main purpose of the object is to provide methods for authenticating with AWS Secret Manager
        and Google's Sheets API.

        Args:
            aws_session: None or already existing boto3 session to be used to get secrets from the AWS secret manager
            aws_credentials: access key id (idx 0), secret key (idx 1) and token (idx 2) to authenticate with AWS
        """
        
        # Reseting package level google credentials object and user's token
        fc.gcreds = None
        fc.token = None

        self.session = None

        if not aws_session and len(aws_credentials) < 1:
            raise AuthenticationError("Can not instantiate auth object without AWS credentials")

        # AWS session (boto3)
        if aws_session is None:
            self.session = boto3.session.Session(
                aws_access_key_id=aws_credentials[0],
                aws_secret_access_key=aws_credentials[1],
                aws_session_token=aws_credentials[2]
            )
        else:
            self.session = aws_session


    def get_ggl_creds(self, aws_secret_name: str, aws_region_name:str):
        """
        This method will use the instantiated object and its connection to AWS in
        order to utilize the returned as secret client_id for the Jupyter notebook 
        application that exports the customers' reports to Googles Sheets files.
        This method will only works for aws secrets' values stored as string, but 
        in the format of Google's credentials.json file. 

        Args:
            aws_secret_name: the name of the secret in the AWS Secret Manager
            aws_region_name: the AWS region name

        Returns:
            object: the required by Google's Sheets API credentials

        Raises:
              AuthenticationError: when the credentials are not acquired by any reason
              Exception: general exceptions
        
        """

        try:
            if fc.token:
                fc.gcreds = Credentials.from_authorized_user_info(fc.token, SCOPES)
            
            # If there are no (valid) credentials available, let the user log in.
            if not fc.gcreds or not fc.gcreds.valid:
                if fc.gcreds and fc.gcreds.expired and fc.gcreds.refresh_token:
                    fc.gcreds.refresh(Request())
                else:
                    try:

                        fc.gcreds = self._get_trans_creds(aws_secret_name, aws_region_name)

                    except Exception as err:
                        exc_type, exc_obj, exc_tb = sys.exc_info()
                        print(f"[get_ggl_creds] Error: {err} raised at line {exc_tb.tb_lineno}")
                        raise err

                # Save the credentials for the next run
                fc.token = fc.gcreds.to_json()

            if fc.gcreds is None:
                raise AuthenticationError("User authentication failed!")
        
        except Exception as err:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[get_ggl_creds] Error: {err} raised at line {exc_tb.tb_lineno}")


    def _get_trans_creds(self, aws_secret_name:str, aws_region_name:str) -> object:
        """
        Internal helper function to convert the aws secret value into expected Google format

        Args:
            aws_secret_name: the name of the secret in the AWS Secret Manager
            aws_region_name: the AWS region name

        Returns:
            object: the credentials to be used with Google Sheets API in expected Google format

        Raises:
            Exception: general exception

        """

        try:
            clientid_dict = self._get_secret_from_awssm(aws_secret_name, aws_region_name)
            client_config = ast.literal_eval(json.dumps(clientid_dict))

            flow = InstalledAppFlow.from_client_config(client_config=client_config, scopes=SCOPES)

            credentials = flow.run_local_server(open_browser=True)

        except Exception as err:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[get_trans_creds] Error: {err} raised at line {exc_tb.tb_lineno}")
            raise err
        
        return credentials 


    def _get_secret_from_awssm(self, secret_name: str, region_name: str) -> dict:
        """
        Using the AWS session provided to the object, connects to the AWS Secret manager and returns 
        a dictionary where: the key is the secret id and the value is the value of the secret.
        visit the AWS docs: https://aws.amazon.com/developer/language/python/
        
        Args:
            secret_name: the name of the secret 
            region_name: the name of the region

        Returns:
            dict: the key is the secret id and the value is the value of the secret.
        """
        
        secret_name = secret_name 
        region_name = region_name 

        # Create a Secrets Manager client
        client = self.session.client(
            service_name='secretsmanager',
            region_name=region_name
        )

        secret_group = secret_name.split("/")

        try:
            get_secret_value_response = client.get_secret_value(
                SecretId=secret_name
            )

            # Decrypts secret using the associated KMS key.
            secret = get_secret_value_response['SecretString']
            secret_dict = ast.literal_eval(secret)

            # Extract the secret value only (w/o the key)
            secret_str_value = secret_dict.get(secret_group[0]) 
            secret_value_dict = ast.literal_eval(secret_str_value)

        except ClientError as e:
            # For a list of exceptions thrown, see
            # https://docs.aws.amazon.com/secretsmanager/latest/apireference/API_GetSecretValue.html
            raise e

        except Exception as err:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[get_secret_from_awssm] Error: {err} raised at line {exc_tb.tb_lineno}")
            raise err
        
        return secret_value_dict 




import sys
from fore_cloudreach.errors import ReadingMapFileError, ReportCreationError, AuthenticationError

from googleapiclient.discovery import build
from googleapiclient.errors import HttpError

class Template:
    """
        The `Template` class will represent Google Sheets spreadsheet (template) used to create and feed GS report file with 
        prepared data data.
    """

    def __init__(self, template_id: str, dm: bool) -> None:

        self.debug_mode = dm
        
        if template_id == "":
            raise Exception("The template id is required")
        self.id = template_id


    def new_from_template(self, creds: object, spreadsheet_id: str, report_name: str) -> object:
        """ 
        **Duplicates** the first tab sheet of the given template object (file)
            
        This must be called before ingesting the monthly report data for each customer.
        It will return the new tab sheet id the is required for the ingestion process.

        Args:
            creds: the current user credentials that has permissions to edit the file
            spreadsheet_id: the Google Sheets global unique id of the file that will get the duplcated tab sheet
            report_name: the name of the duplicated sheet

        Returns:
            object: the feedback from the Google's Sheets API

        Raises:
            
        """

        requests = []
        requests.append({
            # Duplicates the contents of a sheet. # Duplicates a sheet.
            "duplicateSheet": { 
                # The zero-based index where the new sheet should be inserted. The index of all sheets after this are incremented.
                "insertSheetIndex": 1,
                # If set, the ID of the new sheet. If not set, an ID is chosen. If set, the ID must not conflict with any existing sheet ID. If set, it must be non-negative. 
                "newSheetId": None, 
                # The name of the new sheet. If empty, a new name is chosen for you.
                "newSheetName": report_name, 
                # The sheet to duplicate. If the source sheet is of DATA_SOURCE type, its backing DataSource is also duplicated and associated with the new copy of the sheet. No data execution is triggered, the grid data of this sheet is also copied over but only available after the batch request completes.
                "sourceSheetId": 0,
            } 
        })

        # API docs: https://googleapis.github.io/google-api-python-client/docs/dyn/sheets_v4.spreadsheets.html#batchUpdate
        try:

            service = build('sheets', 'v4', credentials=creds)

            # Call the Sheets API for duplicating the first sheet (index=0) of the spreadsheet
            response = service.spreadsheets().batchUpdate(
                spreadsheetId=spreadsheet_id, 
                body={'requests':requests}
                ).execute()
            
            # [dev]: extract the sheetId from the response object!

            # The `response` object sample:
            # {'spreadsheetId': '1Ni5B7zolPKD2J0CXQSaLec31AXw8HpcdM_xPJKPILRw', 
            #  'replies': [
            #     {'duplicateSheet': 
            #        {'properties': {'sheetId': 431392078, 'title': '11-2022', 'index': 1, 'sheetType': 'GRID', 'gridProperties': {'rowCount': 1000, 'columnCount': 26}}}}]}
            return response

        except HttpError as err:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[new_from_template] Error: {err} raised at line {exc_tb.tb_lineno}")
            raise err


    def read_map(self, creds, readrange) -> list:
        """ 
        *read_map* will read the provided range from the mapping file

        Args:
            creds     -- authentication credentials
            readrange -- the range where the mapping is stored
        """
        if not creds:
            raise AuthenticationError("Missing credentials object when try to read the customer mapping file.")
         
        try:
            service = build('sheets', 'v4', credentials=creds)
            
            # Call the Sheets API
            sheet = service.spreadsheets()

            result = sheet.values().get(spreadsheetId=self.id,
                                        range=readrange).execute()

            values = result.get('values', [])

            if not values:
                print('No data found.')
                return []

            range_width = len(values[0])
            map_titles = []
            cstmmap = []

            for idx, row in enumerate(values):
                
                if idx == 0:
                    for i in range(range_width):
                        map_titles.append(row[i])
                    continue
                
                map_row = []
                for j in range(range_width):
                    map_row.append(row[j])

                cstmmap.append(map_row)
            
            return cstmmap

        except HttpError as err:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[read_map] Error: {err} raised at line {exc_tb.tb_lineno}")
            raise ReadingMapFileError(str(err.content))


    def write_report_values(self, creds: object, data: list) -> object:
        """ 
            *write_repost_values* write the report in the object's spreadsheet

            source: https://googleapis.github.io/google-api-python-client/docs/dyn/sheets_v4.spreadsheets.values.html#batchUpdate

            keyword arguments:

            * creds -- is the credentials object with permissions to the spreadsheet
            * data  -- is a list of objects; Each object is of the type below:

        ```
        .. highlight:: python
        .. code-block:: python

            { #### Data within a range of the spreadsheet.
                ##### The major dimension of the values. For output, if the spreadsheet data is:
                ##### `A1=1,B1=2,A2=3,B2=4`, then requesting `range=A1:B2,majorDimension=ROWS` will return `[[1,2],[3,4]]`, 
                ##### whereas requesting `range=A1:B2,majorDimension=COLUMNS` will return `[[1,3],[2,4]]`. 
                ##### For input, with `range=A1:B2,majorDimension=ROWS` then `[[1,2],[3,4]]` will set `A1=1,B1=2,A2=3,B2=4`. 
                ##### With `range=A1:B2,majorDimension=COLUMNS` then `[[1,2],[3,4]]` will set `A1=1,B1=3,A2=2,B2=4`. 
                ##### When writing, if this field is not set, it defaults to ROWS.

                "majorDimension": "A String",

                ##### The range the values cover, in [A1 notation](/sheets/api/guides/concepts#cell). 
                ##### For output, this range indicates the entire requested range, even though the values will exclude 
                ##### trailing rows and columns. When appending values, this field represents the range to search for a table, 
                ##### after which values will be appended.

                "range": "A String",

                ##### The data that was read or to be written. This is an array of arrays, the outer array representing all 
                ##### the data and each inner array representing a major dimension. Each item in the inner array corresponds 
                ##### with one cell. 
                ##### For output, empty trailing rows and columns will not be included. 
                ##### For input, supported value types are: bool, string, and double. Null values will be skipped. 
                ##### To set a cell to an empty value, set the string value to an empty string.
                
                "values": [ 
                    [
                        "",
                    ],
                ],
            },
        ```
        """

        if self.debug_mode:
            print(f"**[fore] Length of data list to export to Google Sheets:{len(data)}**")
            print(f"**[fore] Data list content:{data}**")
        else:
            print(f"**[fore] Debug mode logging is: {self.debug_mode}**")

        try:

            service = build('sheets', 'v4', credentials=creds)

            request =   { # The request for updating more than one range of values in a spreadsheet.
                            # The new values to apply to the spreadsheet.
                            "data": data,

                            # Determines if the update response should include the values of the cells that were updated. 
                            # By default, responses do not include the updated values. The `updatedData` field within each of 
                            # the BatchUpdateValuesResponse.responses contains the updated values. 
                            # If the range to write was larger than the range actually written, the response includes all values 
                            # in the requested range (excluding trailing empty rows and columns).
                            "includeValuesInResponse": False,

                            # Determines how dates, times, and durations in the response should be rendered. This is ignored 
                            # if response_value_render_option is FORMATTED_VALUE. The default dateTime render 
                            # option is SERIAL_NUMBER.
                            "responseDateTimeRenderOption": None,

                            # Determines how values in the response should be rendered. The default render option is FORMATTED_VALUE.
                            "responseValueRenderOption": "FORMATTED_VALUE",

                            # How the input data should be interpreted. The values `USER_ENTERED` is seen in an [example](https://developers.google.com/sheets/api/guides/values#writing_to_a_single_range)
                            "valueInputOption": "USER_ENTERED", 
                        } 

            response = None

            # Call the Sheets API for duplicating the first sheet (index=0) of the spreadsheet
            response = service.spreadsheets().values().batchUpdate(
                spreadsheetId=self.id,
                body=request,
                x__xgafv=None,
            ).execute()

        except Exception as err:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[write_report_values] Error: {err} raised at line {exc_tb.tb_lineno}")
            raise ReportCreationError(f"Failed to update the spreadsheet! Error {err}")

        return response

    def show_debug_mode(self):
        print(f"The logging debug mode is: {self.debug_mode}")



import sys
from fore_cloudreach.errors import AuthenticationError, ReadingMapFileError, EmptyMapFileError, ReportCreationError
from fore_cloudreach.template import Template

import pandas as pd
import datetime

class Ingester:
    """ 
    Ingester loads data into new customer finance report
        
    In order to function properlly the object instantiated from this class
    will need to have defined mapping file (spreadsheet) ID which introduce
    the mapping of the customers' name/id to their spreadsheet file IDs 
    (URL unique global file ID).
    Each report must be as well cataloged in the tab sheet `Reports Catalog`
    of the same spreadsheet. 
    One Ingester object will work with one mapping file always. In case more
    mapping files have to be used, one object per each must be instantiated.

    Args:
        creds: Google's Sheets API credentials
        mapping_file_id: the URL global unique file id for the customers' mapping file
    """

    def __init__(self, creds, mapping_file_id: str, dm: bool) -> None:
        """
        Instantiate an object from class Ingester

        Args:
            mapping_file_id: the spreadsheet ID of the Google's Sheets file that contains
                            mapping data between the customer and the spreadsheet_id of the file used for 
                            this customer's financial reports.
                            If an empty string is provided, the default value set in the __init__ will be used.
                            In case the default value is an empty string as well, the user will be prompted
                            to enter a value.
        Returns:
            None

        Raises:
                AuthenticationError: missing credentials
                FileNotFoundError: missing ID for the customers mapping file        

        """
        
        self.debug_mode = dm

        if not creds:
            raise AuthenticationError("Missing credentials")

        # Set default value for map_id. In event of an empty string is assigned - the user will be prompted to enter one.
        map_id = "1fL3rZDj8tCP4povb3E2x_WmkqNmfEZIR_KT_Zbz7p0M"
        
        # Both customers_map and reports_catalog are list of lists, where lists are row values

        # customers_map columns: Customer_Name, Customer_ID, Spreadsheet_ID, AWS_Org_ID 
        customers_map = []
        # reports_catalog columns: ID, Report Name
        reports_catalog = []
        
        report_name = ""
        report_id = 0

        if mapping_file_id == "" and map_id == "":
            self.map_id = input("Enter the Google Sheet file ID from the URL for the customers' mapping file:\n")

        if mapping_file_id != "":
            map_id = mapping_file_id

        if map_id == "":
            raise FileNotFoundError("There is no defined mapping file id!")

        
        try:
            self.creds = creds
            self._read_map_file(self.creds, map_id)

        except AuthenticationError as auterr:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"Ingester:[__init__] Error: {auterr} raised at line {exc_tb.tb_lineno}")
            raise auterr

        except ReadingMapFileError as rmferr:
            rmferr.__str__ = "Unable to read the customers to spreadsheets mapping file!"
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"Ingester:[__init__] Error: {rmferr} raised at line {exc_tb.tb_lineno}")
            raise rmferr
        
        except EmptyMapFileError as emferr:
            emferr.__str__ = "Unable to use customers map or Reports Catalog!"
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"Ingester:[__init__] Error: {emferr} raised at line {exc_tb.tb_lineno}")
            raise emferr

    
    def load_from_df(self, customer: str, df: pd.DataFrame) -> object:
        """ 
        *load_from_df* will receive a pandas data frame with report's data to load in the customer's data spreadsheet.
        This method is sorta "driver" for the report's data for each customer.
        It will be called from reports data exporter ran in the Jupyter notebook.
            
        Args:
            customer: name or id as string for the customer for which the data is to be loaded
            df: the pandas dataframe contaning the data to load

        Returns:
            object: An object as returned by the Google's Sheets API with the result of the updated values.

        Raises:
            FileNotFoundError: Raises an exception.
            ValueError: ...
            ReportCreationError: ...
        """

        # TODO [dev] plan:

        # 1. Identify customer (name, id?) etc.
        # 2. Read from settings file about mapped customers - spreadsheet_id.
        # 3. Instantiate new Template object with the received spreadsheet_id in p.2.
        # 4. Call the function `new_from_template` to duplicate the first tab sheet
        # 5. Cycle through the data frame `df` and load the customer report's data in the new tab sheet created at p.4
        # 6. Feed the status back to the calling procedure
        
        if customer == "":
            raise ValueError("Unidentifiable customer!")
        
        if self.debug_mode:
            print(f"**[fore] Data types in the pandas data frame to be imorted: {df.dtypes}**")
            print(f"**[fore] Records in the panadas data frame to be imported: {len(df.index)}**")
        else:
            print(f"**[fore] Debug mode logging is: {self.debug_mode}**")
        
        try:

            # ------- step-2 --------- 
            spreadsheet_id = self._get_sprd_id(customer)

            if spreadsheet_id == "":
                raise FileNotFoundError(f"Can not find spreadsheet ID for customer {customer}")

            # ------- step-3 ---------                
            target_file = Template(template_id=spreadsheet_id, dm=self.debug_mode)
            month_year = datetime.datetime.now().strftime('%Y-%m-%d-%H%M%S')

            # -------- step-4 --------
            resp = target_file.new_from_template(self.creds, spreadsheet_id, month_year)
            sheetAttributes = self._parse_new_sheet_id(resp)
            if sheetAttributes[0] == 0 or sheetAttributes[1] == "":
                raise ValueError("Unable to identify the report's taget tab sheet id.")

            # -------- step-5 --------
            # Load the panda Data Frames into the Google's Sheet
            
            values = []
            
            for key, val in df.iterrows():
                tmp_val = []
                for i, v in val.items():
                    tmp_val.append(v)
                values.append(tmp_val)

            data = {
                "majorDimension": "ROWS",
                "range": f"{sheetAttributes[1]}!A2:ZZ999",
                "values": values,
            }
                        
            result = target_file.write_report_values(creds=self.creds, data=data)
            if result is None:
                raise RuntimeError({"module": f"{self.__name__}", "Message": f"Unsuccessful import from df"})

            return result

        except FileNotFoundError as err:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[load_from_df] Error while importing data frame: {err} raised at line {exc_tb.tb_lineno}")
            raise err
            
        except ValueError as verr:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[load_from_df] Error while importing data frame: {verr} raised at line {exc_tb.tb_lineno}")
            raise verr

        except ReportCreationError as rerr:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[load_from_df] Error while importing data frame: {rerr} raised at line {exc_tb.tb_lineno}")
            raise rerr

    def load_report(self, customer: str, data: any) -> object:
        """
        This will load a report using the new concept of having 1 customer's Google Sheets File with many
        different reports within it split in tab sheets per YYYY-MM. That should involve as well `Reports Catalog`
        tab sheet from the Customers_mapping file.

        Args:
            customer: the name (or ID) of a customer to which the report data will be loaded. It must be one of the known customers in the list
            data: one of the types:
                - <str> as file name of a CSV file (will be used to determine the report name as well) or 
                - <object> as pandas Data Frame object

        Returns:
            object: An object as returned by the Google's Sheets API with the result of the updated values.

        Raises:
            ValueError:
            FileNotFoundError:

        """

        # [dev] plan:
            # 1. Identify customer (name, id?) self.customers_map
            # 2. Read from settings file about mapped customers - spreadsheet_id.
            # 3. Read the report name (panda DF.index or CSV file name) from Reports Catalog and get the report ID.
            #   3.1 Identify the data source:
            #       - [pandas Data Frame]:
            #       - TODO [dev]: [customer's folder with CSV file] 
            #       - [single CSV file]
            # 4. Instantiate new Template object with the received spreadsheet_id in p.2
            # 5. Call the function `new_from_template` to duplicate the first tab sheet and name it after the pattern:
            #    (name pattern: `<ReportID>-<YYYY-MM><DD-HHMMSS>`)
            # 6. 
            # 7. Cycle through the data source(s) `df` and load the customer report's data in the new tab sheet
            # 8. Feed the status back to the calling procedure

        if customer == "":
            raise ValueError("Unidentifiable customer!")
        
        try:

            # ------- step-2 --------- 
            spreadsheet_id = self._get_sprd_id(customer)

            if spreadsheet_id == "":
                raise FileNotFoundError(f"Can not find spreadsheet ID for customer: {customer}")

            # ------- step-3 ---------
            self.report_name = ""
            self.report_id = 0
            data_type = ""

            if type(data) is str:
                path_fragments = data.split("/")
                fn = path_fragments[len(path_fragments)-1]
                self.report_name = fn.lower().split(".csv")[0]
                data_type = "csv"

            elif type(data) is pd.DataFrame:
                idx = data.index
                self.report_name = idx.name.lower().replace(" ", "_")
                data_type = "df" 

            else:
                print(f"unsupported data type: {type(data).__name__}")
                raise TypeError({"Error": TypeError.__str__, "Message": f"Unsupported data type: {type(data).__name__}"})

            if self.report_name == "" or data_type == "":
                raise ValueError({"Message": "Missing report name or type!"})

            for report in self.reports_catalog:
                if report[1] == self.report_name:
                    self.report_id = report[0]

            if self.report_id == 0:
                raise ResourceWarning({"Message": f"Report {self.report_name} not found in Reports Catalog"})
            
            # -------- step-4 ----------
            target_file = Template(template_id=spreadsheet_id, dm=self.debug_mode)
            month_year = datetime.datetime.now().strftime('%Y-%m <%d-%H%M%S>')
            tab_name = f"{self.report_id}-{month_year}"

            # -------- step-5 ----------
            resp = target_file.new_from_template(self.creds, spreadsheet_id, tab_name)
            sheetAttributes = self._parse_new_sheet_id(resp)
            if sheetAttributes[0] == 0 or sheetAttributes[1] == "":
                raise ValueError({"Message": f"Unable to identify the report's target tab sheet id for {tab_name}!"})

            # -------- step-6 ----------
            # Call the importer method for the respective data type provided
            if data_type == "csv":
                result = self._import_from_csv(target=target_file, sheetAttributes=sheetAttributes, file_name=data)

            elif data_type == "df":
                result = self._import_from_df(target_file, sheetAttributes, df=data)

            if result == None:
                raise RuntimeError({"Module":"[load_report]", "Message": "Error while importing data"})

        except FileNotFoundError as err:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[load_report] Error while loading report: {err} raised at line {exc_tb.tb_lineno}")
            raise err

    def _get_sprd_id(self, cstm: str) -> str:
        """ 
        *_get_sprd_id* will look for preconfigured mapping spreadsheet from where to
        extract the spreadsheet for the given in the argument customer.

        Args:
            cstm: the customer name or id string
        
        Returns:
            str: the spreadsheet's ID as string
        """
        
        spreadsheetId = ""

        for row in self.customers_map:
            if cstm == row[0] or cstm == [1]:
                spreadsheetId = row[2]
                break
        
        return spreadsheetId

    def _read_map_file(self, creds: object, mapid: str) -> None:
        """ 
        *_read_map_file* This method will assign values to the object's properties
        `customers_map` - the customers to spreadsheets map and the `reports_catalog` - 
        the id and report name matrix.

        This method will be executed at object instantiation time and will hold the 
        map and the reports catalog during the object life-cycle time.

        Args:
            creds: Credentials of the current user with permissions to read the file
            mapid: The unique Google Sheets file id with the customers' mapping info

        Returns:
            None:

        Raises:
            EmptyMapFileError: in case the method read_map return empty list
                    
        """

        map_file = Template(template_id=mapid, dm=self.debug_mode)

        try:
            self.customers_map = map_file.read_map(creds=creds, readrange="Map!A1:F100")
            if len(self.customers_map) < 1:
                raise EmptyMapFileError("Can not read or an empty map is returned!")

            self.reports_catalog = map_file.read_map(creds=creds, readrange="Reports Catalog!A1:B100")
            if len(self.reports_catalog) < 1:
                raise EmptyMapFileError("Can not read or an empty map is returned!")

            # print(f"customer map: {self.customers_map}\n\n")
            # print(f"reports catalog: {self.reports_catalog}\n\n")

        except ReadingMapFileError as err:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[_read_map_file] Error while reading map file: {err} raised at line {exc_tb.tb_lineno}")
            raise err

    def _parse_new_sheet_id(self, resp: dict) -> list:
        """ 
        *_parse_new_sheet_id* - get the new sheet id from the response
            
        This internal method will parse the response of the template duplication 
        and will extract the new sheet id and return it as string.        
        """
        
        # sample of the rsponse.
        # { 
        #     'spreadsheetId': '1w2tToHjQI7S8cH9neu1rCG6768IWrylF2rYtvLCSkcA', 
        #     'replies': [
        #         {'duplicateSheet': 
        #             {'properties': 
        #                 {
        #                     'sheetId': 1500007005, 
        #                     'title': '2022-12', 
        #                     'index': 1, 
        #                     'sheetType': 'GRID', 
        #                     'gridProperties': {'rowCount': 1000, 'columnCount': 26}}}}]}
        

        try:
            # List of attributes (id, name) of the new created tab sheet
            response = [0, ""]
            response[0] = resp['replies'][0]['duplicateSheet']['properties']['sheetId']
            response[1] = resp['replies'][0]['duplicateSheet']['properties']['title'] 
       
        except:
            Exception("Failed to extract the new sheetId!")
        
        return response

    def _import_from_df(self, target: object, sheetAttributes: list, df: object) -> object:
        """
        This internal method will import data from pandas data frame into the customer's 
        spreadsheet (target)

        Args:
            target: the customer's spreadsheet
            sheetAttributes: a list of attributes of the spreadsheet
            df: the pandas Data Frame with the import data

        Returns:
            object: 
        """
          
        values = []
        columns = []
        report_name = []

        try:
            
            report_name.append([self.report_name])

            for col in df.columns:
                columns.append(col)
            values.append(columns)

            for key, val in df.iterrows():
                tmp_val = []
                for i, v in val.items():
                    tmp_val.append(v)
                values.append(tmp_val)

            data = [
                {
                    "majorDimension": "ROWS",
                    "range": f"{sheetAttributes[1]}!B1",
                    "values": report_name, 
                },
                {
                    "majorDimension": "ROWS",
                    "range": f"{sheetAttributes[1]}!B2:ZZ999",
                    "values": values,
                },
            ]        
            result = target.write_report_values(creds=self.creds, data=data)
            if result is None:
                raise RuntimeError({"module": "_import_from_df", "Message": "Unsuccessful import from df"})

            print(f"successful import as: {result}")
            return result

        except Exception as impdferr:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[_import_from_df] Error while importing form pandas data frame: {impdferr} raised at line {exc_tb.tb_lineno}!")
            raise impdferr

    def _import_from_csv(self, target: object, sheetAttributes: list, file_name: str) -> object:
        """
        imort csv file into pandas data frame and then invoke `_import_from_df` with the new df object
        
        Args:
            target: the customer's spreadsheet
            sheetAttributes: a list of attributes of the spreadsheet
            file_name: the CSV file name including the reference path to the it
        
        Returns:
            object: the API feedabck for the imort
        
        """
        try: 
            # TODO [dev]: For reacher functionality of the import from CSV, the read_csv function needs
            # to be configured with additional config object in extended implementation 
            data = pd.read_csv(file_name, index_col=False, keep_default_na=False, na_filter=False)

            idx = data.index
            idx.name = self.report_name

            result = self._import_from_df(target=target, sheetAttributes=sheetAttributes, df=data)

            if result is None:
                raise RuntimeError({"module": "_import_from_csv", "Message": "Unsuccessful import from df"})

            print(f"successful import as: {result}")
            return result
        
        except Exception as errimpcsv:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            print(f"[_import_from_csv] Error while importing CSV file: {errimpcsv} raised at line {exc_tb.tb_lineno}")
            raise errimpcsv
    def show_debug_mode(self):
        print(f"The logging debug mode is: {self.debug_mode}")


import unittest
import boto3
import fore_cloudreach as fc

from fore_cloudreach import Auth
from fore_cloudreach import Ingester
from fore_cloudreach import Template
from fore_cloudreach import AuthenticationError, EmptyMapFileError, ReadingMapFileError, ReportCreationError


class TestAuth(unittest.TestCase):
    """
    Test suit for the class Auth
    
    Testing the object instantiation (__init__):
        - test_0.. - test_2..
    
    Testing `get_ggl_creds` method:
        - test_3.. -  

    Testing `_get_trans_creds` method:
        - ...

    Testing `_get_secret_from_awssm` method:
        - ...
    """

    def test_0_Auth_misses_aws_credentials(self):
        print("testing Auth class to raise an AuthenticationError ...")
        with self.assertRaises(AuthenticationError) as context:
            Auth(None, [])
    
    def test_1_Auth_resets_globals(self):
        print("testing Auth class to reset global vars at init time ...")
        dm1 = {"1":"A"}
        dm2 = {"2":"B"}
        dm3 = {"3":"C"}

        fc.gcreds = dm1
        fc.token = dm2

        auth = Auth(dm3, ["100", "101"])
        self.assertIsNone(fc.gcreds)
        self.assertIsNone(fc.token)

    def test_2_build_aws_session(self):
        print("testing Auth build aws boto3 session...")
        auth = Auth(None, ["1", "2", "3"])
        self.assertIsInstance(auth.session, boto3.Session)

    def test_3_assert_exception_on_empty_params(self):
        pass
        
if __name__ == '__main__':
    unittest.main()
